<?php

require 'inc.functions.php';

if ( file_exists('_version.phps') )
{
	include_once('_version.phps');
}

error_reporting(E_ALL|E_NOTICE);
define ('SUCCEEDED', "SUCCEEDED");
define ('SUCCEEDED_WITH_WARNING', "SUCCEEDED_WITH_WARNING");
define ('FAILED', "FAILED");

define ('OUTPUT_FILENAME_EXTENSION', 'utf');
define ('OUTPUT_FILENAME_EXTENSION_SCRIPT', 'script');
if ( !defined('WIKI_NS') )
	define ('WIKI_NS', 'Tomoyo After CS');

class Leave extends Exception {}
class LocalException extends Exception {}
class NotEnoughInputException extends LocalException {}
class ContinuableLocalException extends LocalException {}

$stderr = STDERR;

//#####################################################################################################################

$asmscript_inputs = $_SERVER['argv'];
$is_recursive_mode = FALSE;

$asmscript_inputs = array_slice($_SERVER['argv'], 1);
if ( count($asmscript_inputs) > 0 && $asmscript_inputs[0] == '-R' )
{
	$is_recursive_mode = TRUE;
	array_shift($asmscript_inputs);
}
$asmscript_inputs = expand_argspec($asmscript_inputs, $base_dir = '.');

// No matching file
if ( count($asmscript_inputs) < 1 )
{
	usage(sprintf("No matching input\r\nphp %s [-R] [<filespec.*txt>...|<filespec>...]\r\n -R  Recurse into chained scenarii", $_SERVER['argv'][0]));
	exit;
}
$succeeded = 0;
$succeeded_with_warning = 0;
$failed = 0;

// Depth-first traversal:
// Process script <i> specified on command line plus all its children (if applicable)
// Then process script <i+1>

$process_state = array();

$global_state = array(
	&$process_state,
	&$succeeded,
	&$succeeded_with_warning,
	&$failed,
	&$asmscript_inputs,

);

// Process scripts, $asmscript_inputs may be changed by process_input_path() if recursive mode is enabled

while ( count($asmscript_inputs) > 0 )
{
	$asmscript_filepath = array_shift($asmscript_inputs);
	if ( !isset($process_state[basename($asmscript_filepath)]) )
	{
		  // when returning from recursion
		process_input_path($asmscript_filepath, $stderr, $is_recursive_mode, $global_state);
	}
}


fprintf(STDERR, "Analyze: %d inputs, %d SUCCESS (%d with WARNING), %d FAILURE\r\n",
	$succeeded + $failed,
	$succeeded,
	$succeeded_with_warning,
	$failed);

//#####################################################################################################################
// end driver program


function create_initial_scriptinfo($lines_array_elt)
{
	return array(0, $lines_array_elt);
}

function is_label_def($value, &$matches)
{
	// may be empty brackets
	return preg_match('/_ZY.+\\((LV[^)]+)/', $value, $matches);
}

function is_dialogue_line($value, &$matches)
{
	// WARNING: modify separator token (comma) in PS2 executable
	return preg_match('/_ZM.+\\(([^)]*)/', $value, $matches) || preg_match('/_MSNL.+,([^)]*)/', $value, $matches);
}

function is_sel_line($value, &$matches)
{
	return preg_match('/(.\\[[0-9]+\\])=SEL\\(([^)]+)/', $value, $matches);
}

function _MXNP($charnum)
{
	global $scene_personae;
	$scene_personae[$charnum] = TRUE;
	return $charnum;
}

function get_actor($charnum)
{
	static $chars = array(
		'朋也',
		'智代',
		'声',
		'鷹文',
		'親方',
		'初老の男',
		'智代・朋也',
		'女の子',
		'とも',
		'先生',
		'女性',
		'初老の女性',
		'主婦Ａ',
		'主婦Ｂ',
		'管理人',
		'母親',
		'クマ',
		'パンダ',
		'主婦',
		'河南子',
		'男',
		'影',
		'子供',
		'キャプテン',
		'部員',
		'有子',
		'書道の先生',
		'村人',
		'老婆',
		'少女',
		'医師',
		'男子生徒',
		'小川',
		'女生徒',
		'直幸',
		'側近',
		'父親',
		'医者',
		'老人',
	);
	return $chars[$charnum];
}



function process_input_path($asmscript_filepath, $stderr, $is_recursive_mode, $global_state)
{
	$process_state =& $global_state[0];
	$succeeded =& $global_state[1];
	$succeeded_with_warning =& $global_state[2];
	$failed =& $global_state[3];
	$asmscript_inputs =&  $global_state[4];

	$process_state[basename($asmscript_filepath)] = TRUE;  // being processed

	$status = SUCCEEDED;

	list($base, $ext) = break_filename($asmscript_filepath);
	$scriptname = basename($base);

preg_match('/(.+)[0-9]+$/', $scriptname, $matches);
$prefix = $matches[1];

fprintf($stderr, "* %s\r\n", basename($asmscript_filepath));


	try
	{
		$inputdata = @file($asmscript_filepath);
		if ( $inputdata === FALSE )
		{
			fprintf($stderr, "%s: Cannot open specified file\r\n", basename($asmscript_filepath));
			throw new Leave();
		}

		if ( $ext === OUTPUT_FILENAME_EXTENSION )
		{
			fprintf($stderr, "%s: Specified file does not have an accepted extension\r\n", basename($asmscript_filepath));
			throw new Leave();
		}

		$scriptinfo = array_map('create_initial_scriptinfo', $inputdata);
		unset($inputdata);

		$out = "\xEF\xBB\xBF".
		"{{ScriptTableBegin}}\r\n".
		"{{**|-9999|Localization file for $scriptname}}\r\n".
		"\r\n";

		$dialogue_lines = array();
		$speakers = array();
		global $scene_personae;
		$scene_personae = array();
		$dependents = array();
		$labels = array();
		$jumps = array();
		$events = array();

		$if_stack = array();
		$current_actor = '';
		$condition = '';
		$nextdlg = '';

		try
		{
			foreach ( $scriptinfo as $numline => &$lineinfo )
			{
				if ( is_label_def($lineinfo[1], $matches) )
				{
					$label = trim($matches[1]);
					if ( isset($labels[$label]) && $labels[$label] != $numline )
					{
						fprintf($stderr, "%s: Label %s defined more than once! (First @ %d, now @ %d)\r\n", basename($asmscript_filepath), $label, $labels[$label] + 1, $numline + 1);
						throw new Leave();
					}
					$labels[$label] = $lineinfo[0];
				}
				elseif ( preg_match('#_IFJP.*\\*(LV[^\\;]+)#', $lineinfo[1], $matches) )
				{
					$label = trim($matches[1]);

					// Is target label already registered?
					if ( isset($labels[$label]) )
					{
						$labellineinfo = $scriptinfo[$labels[$label]];
//						fprintf($stderr, " ==> Jump\t\t[%s] @ [%d --> %d]%s, current level = (%d), new level = :%d:\r\n",
//						  $label, $numline + 1, $labels[$label],
//						  $labels[$label] < $numline ? ' (BACKWARDS)' : '',
//						  $lineinfo[0], $labellineinfo[0]
//						);

						// forward "normal" if-block
						if ( $labels[$label] > $numline )
						{
//							fprintf($stderr, "  |__ Found [%d --> %d]\r\n", $numline + 1, $labels[$label] + 1);

							// all lines in-between are "indented"
							for ( $k = $numline + 1; $k < $labels[$label]; ++$k )
								++$scriptinfo[$k][0];
						}

						// register jump
						$jumps[$numline] = $labels[$label];
					}
					else  // Look for missing label
					{
//						fprintf($stderr, " ==> Jump\t\t[%s] @ %d, searching location for new label, current level = (%d)...\r\n", $label, $numline + 1, $lineinfo[0]);
						$labeldef = FALSE;
						for ( $j = $numline + 1; $j < count($scriptinfo); ++$j )
						{
							$matches = array();
							if ( is_label_def($scriptinfo[$j][1], $matches) )
							{
								$newlabel = trim($matches[1]);

								// To avoid having to search again when referenced, define new label
								if ( !isset($labels[$newlabel]) )
									$labels[$newlabel] = $j;

								if ( $newlabel == $label )
								{
									// FOUND at line $j
									$labeldef = TRUE;
//									fprintf($stderr, "  |__ Found [%d --> %d]\r\n", $numline + 1, $j + 1);

									// all lines in-between are "indented"
									for ( $k = $numline + 1; $k < $j; ++$k )
										++$scriptinfo[$k][0];

									// register jump
									$jumps[$numline] = $j;
									break;
								}
							}
						}

						if ( $labeldef === FALSE )
						{
							fprintf($stderr, "%s: Reference to undefined label %s @ %d\r\n", basename($asmscript_filepath), $label, $numline + 1);
							throw new Leave();
						}
					}
					$events[$numline] = array('jmp', $labels[$label], NULL, $condition, $switchstart = FALSE);

				}
				elseif ( preg_match('#IF_+\\((.+)\\);#', $lineinfo[1], $matches) )  // _IF__ / _EIF_
				{
					$condition = $matches[1];
				}
				elseif ( preg_match('#_MXNP\\(([0-9]+)\\);#', $lineinfo[1], $matches) )
				{
					eval($matches[0]);
					$current_actor = get_actor(intval($matches[1]));
				}
				elseif ( is_dialogue_line($lineinfo[1], $matches) )
				{
					if ( defined('ACTOR_INLINE') )
					{
						if ( $matches[1] == '【' )
						{
							// do not register event, append for next
							$nextdlg = '【!!';
						}
						else
						{
							$matches[1] = $nextdlg. $matches[1];
							$nextdlg = '';
						}
						if ( preg_match('/^【([^】]+)】(.*)/u', $matches[1], $newmatches) )
						{
							$current_actor = $newmatches[1];
							$matches[1] = $newmatches[2];
							global $scene_personae;
							$scene_personae[$current_actor] = TRUE;
						}
					}
					$dialogue_lines[$numline] = array($numline, $matches[1]);
					$speakers[$numline] = $current_actor;
					$events[$numline] = array('dlg');

					$current_actor = '';
				}
				elseif ( is_sel_line($lineinfo[1], $matches) )
				{
					$events[$numline] = array('sel', $matches[1], $matches[2]);
				}
				elseif ( preg_match('#_STTI."([^"]*)#', $lineinfo[1], $matches) /*&& $matches[1] != ''*/ )
				{
					$dialogue_lines[$numline] = array($numline, $matches[1]);
					$speakers[$numline] = 'Set Title';
					$events[$numline] = array('dlg');

					$current_actor = '';
				}
				elseif ( preg_match('#_JUMP\\(([0-9]+)\\);#', $lineinfo[1], $matches) )
				{
					$events[$numline] = array('scn', $matches[1]);
				}
				elseif ( strpos($lineinfo[1], '_FRET;') !== FALSE )
				{
					$events[$numline] = array('frt');
				}
				elseif ( strpos($lineinfo[1], '_RTM;') !== FALSE )
				{
					$events[$numline] = array('rtm');
				}
			}

			//########################################################################################################
			//########################################################################################################
			//########################################################################################################
			// Processed all lines, consistency check

			$tainted = false;
			foreach ( $jumps as $from => $to )
			{
				if ( $scriptinfo[$from][0] != $scriptinfo[$to][0] )
				{
					fprintf($stderr, "%s: Inconsistent jump [%d (%d)  -->  %d (%d)], potential error!\r\n", basename($asmscript_filepath), $from, $scriptinfo[$from][0], $to, $scriptinfo[$to][0]);
					$tainted = true;
				}
			}
			if ( $tainted )
				throw new Leave();

			if ( count($scene_personae) > 0 )
			{
				$out .= "{{//|-2000||Dramatis personae}}\r\n";

				$start_wikimarkup = -2000;

				foreach ( array_keys($scene_personae) as $persona_id )
				{
					if ( defined('ACTOR_INLINE') )
						$actor = $persona_id;
					else
						$actor = get_actor($persona_id);

					$out .= sprintf("{{**|%d|Actor:'''%s'''}}\r\n", ++$start_wikimarkup, $actor);  // UTF-8
				}
				$out .= "\r\n";
			}


//foreach ( $scriptinfo as $numline => &$lineinfo )
//	echo str_repeat("\t", $lineinfo[0]). $lineinfo[1];
//var_dump($labels);
//exit;

			$i = 0;
			$previous_numline = -1;
			//$maxnumline = end(array_keys($events));
			//foreach ( $events as $numline => $eventinfo )
			while ( count($events) > 0 )
			{
				ksort($events);  // I know, not very efficient
				foreach ( $events as $numline => $eventinfo ) break;
				unset($events[$numline]);
				// multiple events on one line?
				if ( is_array($eventinfo[0]) )
				{
					$events[$numline] = $eventinfo[1];
					$eventinfo = $eventinfo[0];
				}

				if ( $eventinfo[0] == 'scn' )
				{
					$mask = "\r\n".
					"{{**|%04d|⇒ scene: \"[[". WIKI_NS.":SEEN%s.utf|SEEN%s]]\"}}\r\n";
					$out .= sprintf($mask, $i, $eventinfo[1], $eventinfo[1]);
					++$i;
				}
				elseif ( $eventinfo[0] == 'frt' )
				{
					$mask = "\r\n".
					"{{**|%04d|⇒ Function return}}\r\n";
					$out .= sprintf($mask, $i);
					++$i;
				}
				elseif ( $eventinfo[0] == 'rtm' )
				{
					$mask = "\r\n".
					"{{**|%04d|⇒ Return main}}\r\n";
					$out .= sprintf($mask, $i);
					++$i;
				}
				elseif ( $eventinfo[0] == 'dlg' )
				{
					$dialogue_info = $dialogue_lines[$numline];
					$dialogue = $dialogue_info[1];
					$speaker = 	$speaker_e = '';

					if ( $speakers[$numline] != '' )
					{
						$speaker = $speakers[$numline];
						if ( $speaker == 'Set Title' )
							$speaker_e = $speaker;
						else
							$speaker_e = 'EN:'. $speaker;
					}
					$mask = "\r\n".
					"{{EN|%1\$04d|%4\$s|}}\r\n".
					"{{JA|%1\$04d|%3\$s|%2\$s}}\r\n";
					$out .= sprintf($mask, $i, $dialogue, $speaker, $speaker_e);
					++$i;
				}
				elseif ( $eventinfo[0] == 'sel' )
				{
					$choices = str_replace(',', ' {{!}} ', $eventinfo[2]);

					$mask = "\r\n".
					"{{EN|%1\$04d|SEL|}}\r\n".
					"{{JA|%1\$04d|SEL|%2\$s}}\r\n".
					"{{**|%1\$04d|User choice ⇒ flag %3\$s}}\r\n";
					$out .= sprintf($mask, $i, $choices, $eventinfo[1]);
					++$i;
				}
				elseif ( $eventinfo[0] == 'jmp' && !$eventinfo[4] )
				{
					$state = array(&$jumps, &$events, &$i, &$out, &$scriptinfo);
					process_if($numline, $eventinfo, $state);
				}
				elseif ( $eventinfo[0] == 'jds' )
				{
					$out .= sprintf("\r\n".
					"{{**|%04d|↓". str_repeat('　', $eventinfo[3]). " Possible %s '''↑[%04d]'''}}\r\n", $i, $eventinfo[2] ? (' switch case('. $eventinfo[2]). ')' : 'jump', $eventinfo[1]);
					++$i;
				}
				elseif ( $eventinfo[0] == 'jse' )
				{
					$out .= sprintf("\r\n".
					"{{**|%04d|	". str_repeat('＊', $eventinfo[2]). " End of switch '''↑[%04d]'''}}\r\n", $i, $eventinfo[1]);
					++$i;
				}

				$previous_numline = $numline;
			}

			$out .= "\r\n\r\n".
			"{{ScriptTableEnd}}\r\n".
			"\r\n".
			"[[Category:". WIKI_NS. ":Scripts]]\r\n";

			file_put_contents($base. '.'. OUTPUT_FILENAME_EXTENSION, $out);
//			file_put_contents($base. '.'. OUTPUT_FILENAME_EXTENSION_SCRIPT, $outbinscript);

		}
		catch(LocalException $e)
		{
			fprintf($stderr, "%s: Specified file is not a valid input file! %s\r\n", basename($asmscript_filepath), $e->getMessage());
			throw new Leave();
		}

		++$succeeded;
		if ( $status == SUCCEEDED_WITH_WARNING )
		{
			++$succeeded_with_warning;
		}

		$process_state[basename($asmscript_filepath)] = 1;  // processed (but not children)

//if (preg_match('/00154/',  $asmscript_filepath))
//{
//	$k = array_keys($scene_actors);
//	sort($k);ksort($process_state);
//	var_dump($k, $process_state, $dependents);exit;
//}


		if ( $is_recursive_mode && count($dependents) > 0 )
		{
			foreach ( $dependents as $id => $foo )
			{
				$target = dirname($base). sprintf("/$prefix%04d.txt",  $id);
				if ( !isset($process_state[basename($target)]) )
				{
					process_input_path($target,
						$stderr,
						$is_recursive_mode,
						$global_state
					);
				}
			}
		}


	}
	catch(Leave $e)
	{
		++$failed;
		$status = FAILED;
	}

}



function process_if($numline, $eventinfo, $state)
{
	$jumps =& $state[0];
	$events =& $state[1];
	$i =& $state[2];
	$out =& $state[3];
	$scriptinfo =& $state[4];

	$condition = $eventinfo[3];

	$flagexp = FALSE;

	$switch = array();
	// for the moment it's merely an if then
	$switch[] = array($numline, $jumps[$numline]);

	// Simple expression
	if ( preg_match('#^(.\\[[0-9]+\\])[<>=-]+[0-9]+#', $condition, $matches) )
		$flagexp = preg_quote($matches[1], '#');
	else
		$matches = array('', '(flag)');

	$cascade_numline = $jumps[$numline] + 2;  // 0: *Label; 1: EIF; 2: IFJP() -- 2 has event registered

	// Try matching if loops thereafter, matching the condition pattern
	while ( isset($jumps[$cascade_numline]) )
	{
		$condition = $events[$cascade_numline][3];
		// register current & mark as processed
		if ( preg_match("#$flagexp#", $condition) )
		{
			$switch[] = array($cascade_numline, $jumps[$cascade_numline] + 2);
			$events[$cascade_numline][4] = ($switchstart = TRUE);
		}
		else
		{
			break;
		}
		$cascade_numline = $jumps[$cascade_numline] + 2;  // 0: *Label; 1: EIF; 2: IFJP() -- 2 has event registered
	}

	if ( count($switch) > 1 )
	{
		$last = end($switch);
		$insert_switchend_marker_line = $last[1] - 2;
		$insert = array('jse', $i, $scriptinfo[$numline][0] + 1);
		if ( !isset($events[$insert_switchend_marker_line]) )
			$events[$insert_switchend_marker_line] = $insert;
		else
			// Put jump dest info after all others
			$events[$insert_switchend_marker_line] = array($events[$insert_switchend_marker_line], $insert);
	}

	$current_i = $i;

	if ( count($switch) > 1 )
	{
		$out .= sprintf("\r\n".
		"{{**|%1\$04d|	". str_repeat('→', $scriptinfo[$numline][0] + 1). " Switch on flag %2\$s, %3\$d destinations (\"switch case(%2\$s) ↑[%1\$04d]\")}}\r\n", $current_i, $matches[1], count($switch));
		++$i;
	}
	else
	{
		$out .= sprintf("\r\n".
		"{{**|%1\$04d|	". str_repeat('→', $scriptinfo[$numline][0] + 1). " Conditional jump (\"Possible jump ↑[%1\$04d]\")}}\r\n", $current_i);
		++$i;
	}

	$k = 0;
	foreach ( $switch as $dest )
	{
		if ( count($switch) > 1 && $k == count($switch) - 1 )
			break;

		$insert_marker_line = $dest[1] - 1;
		$insert = array('jds', $current_i, count($switch) > 1 ? $matches[1] : FALSE, $scriptinfo[$numline][0] + 1);
		if ( !isset($events[$insert_marker_line]) )
			$events[$insert_marker_line] = $insert;
		else
			// Put jump dest info after all others
			$events[$insert_marker_line] = array($events[$insert_marker_line], $insert);

		++$k;
	}
}